
/* 
   (c) COPYRIGHT URI/MIT 1997,1998
   Please read the full copyright statement in the file COPYRIGH.  

   Authors:
      jhrg,jimg		James Gallagher (jgallagher@gso.uri.edu)
*/

/**
  Extend a Matlab vector/matrix by combining an existing array with new data.
  Combining two vectors of length M and N yields a vector of length M + N.
  Combining two matrices of size r1 x c1 and r2 x c2 yields a matrix of size
  r1 x c1 + c2 and presumes that r1 == r2.

  @author jhrg 3/5/97
*/

/* 
 * $Log: extend.c,v $
 * Revision 1.16.16.1  1999/06/01 15:17:13  jimg
 * Moved include of config_writeval.h to fix warning about ENDIAN redefinition.
 *
 * Revision 1.16  1998/11/30 05:37:36  jimg
 * Added the -T option to control truncation of names longer than 19 characters
 *
 * Revision 1.15  1998/11/25 01:15:23  jimg
 * Identifiers limited to 19 chars. This is a temp fix; will update to include ML 5 later
 *
 * Revision 1.14  1998/11/11 18:49:53  jimg
 * Changed the allocation of variables in add_new_variable() from 5 to 500. Dan
 * found that the loaddods/matlab crash we see on Linux is caused by corruption
 * of this data structure when the data structure was initialized to 5 elements.
 * Currently, Mathworks is looking into the problem since it may be inside their
 * code.
 *
 * Revision 1.13  1998/11/10 03:18:14  jimg
 * Added test stuff. This is fairly hokey...
 *
 * Revision 1.12  1998/07/13 20:31:10  jimg
 * Fixes from the final test of the new build process
 *
 * Revision 1.11  1998/06/09 03:59:32  jimg
 * Removed ugly debugging code.
 *
 * Revision 1.10.4.1  1998/05/15 17:43:52  jimg
 * Removed DODS_DEBUG define
 *
 * Revision 1.10  1998/03/23 22:54:39  jimg
 * Fixed extern_strings(...) so that it now works.
 *
 * Revision 1.9  1998/03/21 00:29:39  jimg
 * Added dummy function for extend_strings(). It prints an error message and
 * returns false.
 *
 * Revision 1.8  1998/03/21 00:19:51  jimg
 * Added some documentation.
 * Added intern_strings().
 *
 * Revision 1.7  1997/10/09 22:20:04  jimg
 * Resolved conflicts in merge of 2.14c to trunk.
 *
 * Revision 1.6  1997/10/04 00:33:47  jimg
 * Release 2.14c fixes
 *
 * Revision 1.4.6.1  1997/09/19 19:11:21  jimg
 * Moved config_writeval.h so that -V4_COMPAT flag works.
 *
 * Revision 1.5  1997/09/10 22:47:58  jimg
 * Added config_writeval.h to includes.
 *
 * Revision 1.4  1997/05/01 22:42:14  jimg
 * Added changes for the `return arguments' feature. See loaddods.c
 *
 * Revision 1.3  1997/03/27 22:55:30  jimg
 * Made rm2cm amd cm2rm globally visible.
 *
 * Revision 1.2  1997/03/07 04:54:09  jimg
 * Fixed comments.
 *
 * Revision 1.1  1997/03/07 04:50:26  jimg
 * Added.
 */

static char id[]={"$Id: extend.c,v 1.16.16.1 1999/06/01 15:17:13 jimg Exp $"};

#include <stdio.h>

#include "config_writeval.h"

#include "mex.h"

#include "defines.h"

#ifdef TEST_LOADDODS
#define mexPutFull(a, b, c, d, e) 0
#define mexPutMatrix(a) 0
#define extend_array(a, b, c, d) 1
#define extend_strings(a, b) 1
#endif

extern Matrix **return_args;	/* see loaddods.c */
extern int num_return_args;
extern int truncate_variable_names;

/** When using the return arguments feature of loaddods this holds the
    current argument number. This is re-initialized by loaddods.c:init()
    every time the loaddods command is run. 

    @see num_return_args
    @see return_args */
int current_arg = 0;

#define SEPARATOR "\n"
#define ISVECTOR(r, c) ((r) == 1 || (c) == 1)
#define MATRICES_MATCH(r1, c1, r2, c2) ((r1) == (r2) || (c1) == (c2))

/** Given #src# in row major order return #dest# in column major order. Note
    that #src# and #dest# are one dimensional arrays which contain two
    dimensional data (i.e., they are the memory which comprise the two
    dimensional array).

    Example: If A = 1 2 3
                    4 5 6
	   then #src# = 1 2 3 4 5 6 (row major order)
	   and #dest# = 1 4 2 5 3 6 (column major order) */

void
rm2cm(double *dest, double *src, int rows, int cols) 
{
    int i, j;

    for (i = 0; i < cols; ++i)
	for (j = 0; j < rows; ++j)
	    dest[i*rows + j] = src[i + j*cols];
}

/**
  Given #src# in column major order, return #dest# in row major order. 

  @see rm2cm
*/

void
cm2rm(double *dest, double *src, int rows, int cols) 
{
    int i, j;

    for (i = 0; i < rows; ++i)
	for (j = 0; j < cols; ++j)
	    dest[i*cols + j] = src[i + j*rows];
}

/** Extend the Matlab variable #name# by adding the elements referenced by
    #pr2#. The data referred to by #pr2# may be a scalar, vector or matrix.
    However, its shape must match that of #name#'s in at least one dimension.
    To extend #name#, the values in #pr2# will be added by catenating so that
    the common dimension retains its size (e.g., if #name# is a 2 x 3 matrix
    and #pr2# is a 2 x 4, then the result will be a 2 x 7). If both the row
    and column sizes are the same #name# will be extended by catenating rows.

    NB: This is only visible within this file.

    @param name The name of the variable to extend (this must already exist
    in the Matlab workspace.
    @param pr2 A reference to data to be added to #name#.
    @param r2 The row count of #pr2#.
    @param c2 The column count of @pr2#. */

#ifndef TEST_LOADDODS
static int
extend_array(char *name, int r2, int c2, double *pr2)
{
    int r1, c1;
    double *pr1, *pr;
    int s1, s2;
    int r, c;
    Matrix *m1 = mexGetMatrixPtr(name);
    Matrix *m2;

    if (!m1) {
	fprintf(stderr, "extend_array: could not find array `%s'\n", name);
	return FALSE;
    }

    r1 = mxGetM(m1);		/* Get the row count of Matrix `m1' */
    c1 = mxGetN(m1);		/* And column count */
    pr1 = mxGetPr(m1);		/* And pointer to real data */
	
    DBG(fprintf(stderr, "m1: %x, r1: %d, c1: %d, pr1: %x\n", m1, r1, c1,\
		pr1));

    if (ISVECTOR(r1, c1) && ISVECTOR(r2, c2)) {
	s1 = MAX(r1, c1);
	s2 = MAX(r2, c2);
	/* Use the `shape' of `m1' to determine the shape of the extended
	   vector. */
	r = (c1 == 1) ? s1 + s2 : 1;
	c = (c1 != 1) ? s1 + s2 : 1;
	m2 = mxCreateFull(r, c, 0);

	pr = mxCalloc(s1 + s2, sizeof(double));
	/* Copy old and new data */
	memcpy(pr, pr1, s1 * sizeof(double));
	memcpy(pr + s1, pr2, s2 * sizeof(double));

	DBG(fprintf(stderr, "vector:\n"));
    }
    else if (MATRICES_MATCH(r1, c1, r2, c2)) {
	s1 = r1 * c1;
	s2 = r2 * c2;
	c = (c1 == c2) ? c1 : c1 + c2;
	r = (c1 == c2) ? r1 + r2 : r1;
	m2 = mxCreateFull(r, c, 0);

	pr = mxCalloc(s1 + s2, sizeof(double));
	
	if (c1 == c2) {
	    /* When columns are equal, extend by catenating rows. Given that
	       the Matlab array is already in column major order and the new
	       data is in row major order, the operation proceeds as:
	       1) Transform the Matlab data to row major order.
	       2) catenate the Matlab and DODS data.
	       3) transform the result back to column major order.

	       E.G.: A is the original Matlab data, B is the new information.
	             A = 1 2 3            B = 7 8 9
		         4 5 6
		     
		     a = 1 4 2 5 3 6      b = 7 8 9

		     step 1: a' = 1 2 3 4 5 6
		          2: c = a' b = 1 2 3 4 5 6 7 8 9
			  3  c' = 1 4 7 2 5 8 3 6 9
			  
		     Which yields: C = 1 2 3
		                       4 5 6
				       7 8 9
	    */
	    double *pr1p = mxCalloc(r1 * c1, sizeof(double));
	    double *prp = mxCalloc(r * c, sizeof(double));

	    cm2rm(pr1p, pr1, r1, c1);

	    memcpy(pr, pr1p, s1 * sizeof(double));
	    memcpy(pr + s1, pr2, s2 * sizeof(double));
	    
	    rm2cm(prp, pr, r, c);

	    mxFree(pr);
	    mxFree(pr1p);
	    /* Don't free pr1 since Matlab manages that memory */

	    pr = prp;		/* ...so that mxSetPr() further down works */
	}
	else {
	    /* When rows are equal, extend by catenating columns. Do this by
	       transforming the row major data to column major and appending.
	       E.G.: A = 1 2 3        B = 7 8
	                 4 5 6            9 10

		     a = 1 4 2 5 3 6  b' = 7 9 8 10

		     c = 1 4 2 5 3 6 7 9 8 10
		      
		     C = 1 2 3 7 8
		         4 5 6 9 10
	    */
	    double *pr2p = mxCalloc(r2 * c2, sizeof(double));

	    rm2cm(pr2p, pr2, r2, c2);
	    
	    memcpy(pr, pr1, s1 * sizeof(double));
	    memcpy(pr + s1, pr2p, s2 * sizeof(double));
	    
	    mxFree(pr2p);
	}

	DBG(fprintf(stderr, "matrix:\n"));
    }
    else {
	fprintf(stderr, "extend_array: can only combine vectors or matrices with equal column dimensions\n");
	return FALSE;
    }

    DBG(fprintf(stderr, "s1: %d, s2: %d\n", s1, s2));
    DBG(fprintf(stderr, "r: %d, c: %d\n", r, c));

    /* set name and data of new matrix */
    mxSetPr(m2, pr);
    mxSetPi(m2, 0);
    mxSetName(m2, name);

    /* Intern */
    mexPutMatrix(m2);

    return TRUE;
}
#endif

/** Given the name of a vector of strings (encoded is the special DODS hack
    way - see intern_string()) append the string data pointed to by #s# to
    it. If no such vector exists, return FALSE, otherwise return TRUE. */

#ifndef TEST_LOADDODS
static int
extend_strings(char *name, char *s)
{
    Matrix *mp = mexGetMatrixPtr(name);
    int ostr_len, nstr_len;
    char *ostr, *nstr;

    if (!mp)
	return FALSE;

    ostr_len = mxGetN(mp) + 1;
    ostr = (char *)mxCalloc(ostr_len, 1);
    mxGetString(mp, ostr, ostr_len);
    
    nstr_len = strlen(s) + 1 + ostr_len + 1; /* +1 for <CR> and +1 for null */
    nstr = (char *)mxCalloc(nstr_len, 1);
    strcpy(nstr, ostr);
    strcat(nstr, SEPARATOR);
    strcat(nstr, s);

    mp = mxCreateString(nstr);
    if (mp)
	mxSetName(mp, name);

    return mp && mexPutMatrix(mp) == 0;
}
#endif

/**
  Maintain a list of variables which are new (to this invocation of loaddods)
  so that we know when to clobber something that exists in Matlab versus when
  to extend. The rule is, extend if is_existing_variable() returns true,
  otherwise, don't extend but do call add_new_variable(). Of course, if the
  user does not want `extension' don't do any of this stuff.

  NB: This is only visible within this file.
*/

static struct {
    int len;
    int size;
    char **name;
} variables = {0, 0, 0};

/** Add the name #name# to the list of new variables read during this run of
    loaddods. 

    NB: This is only visible within this file. */

static void
add_new_variable(char *name)
{
    if (variables.len == variables.size) {
	variables.size += 500;
	if (variables.len == 0)
	    variables.name = (char **)malloc(variables.size);
	else
	    variables.name = (char **)realloc(variables.name, variables.size);
    }

    variables.name[variables.len] = (char *)malloc(strlen(name)+1);
    strcpy(variables.name[variables.len++], name);
}

/** Is the variable #name# in the set of variables read during this
    invocation of loaddods? 

    NB: This is only visible within this file.

    @return TRUE if #name# is on the list of current variables, FALSE
    otherwise. */

static int
is_existing_variable(char *name)
{
    int i = 0;
    while (i < variables.len)
	if (strcmp(name, variables.name[i++]) == 0)
	    return TRUE;

    return FALSE;
}

static char *
enforce_name_limits(char *name)
{
    if (truncate_variable_names && (strlen(name) > 19))
	name[19] = '\0';

    return name;

}

/** Wipe clean the list of variable names read so far by loaddods. Normally
    this would be run at the start of each run of loaddods. */

void
clear_existing_variables()
{
    variables.len = 0;
}

/** Return the next name in the list of variable names. Return NULL when there
    are no more names. If FIRST is true, start over from the front of the
    list. */

char *
get_variable_name(int first)
{
    static int item = 0;

    if (first)
	item = 0;

    if (item >= variables.len)
	return NULL;
    else
	return variables.name[item++];
}

/**
   Add into the Matlab workspace a Float variable #name# with values #yp#.
   The variable may be a scalar, vector or matrix - this is determined by the
   row and column sizes given by #r# and #c#. If #extend# is true, then if
   a variable with #name# already exists in the workspace add these values to
   it rather than replacing it with these values.

   NB: When using the return arguments feature of loaddods #extend# has no
   effect (because return args always creates new values and passes them back
   to the caller instead of interning them directly in the Matlab workspace.

   @return TRUE if successful, FALSE otherwise. */
   
int
intern(char *name, int r, int c, double *yp, int extend)
{
    name = enforce_name_limits(name);

    DBG(fprintf(stderr, "entering intern, name: %s, r: %d, c: %d, ext: %d\n",
		name, r, c, extend));

    /* Using the `extend' option and using return arguments is mutually
       exclusive. */

    if (extend) {
	if (is_existing_variable(name)) {
	    if (!extend_array(name, r, c, yp)) {
		fprintf(stderr, "Could not create variable for `%s'\n",
			name);
		return FALSE;
	    }
	}
	else {
	    add_new_variable(name);
	    if (mexPutFull(name, r, c, yp, NULL) != 0) {
		fprintf(stderr, "Could not create variable for `%s'\n",
			name);
		return FALSE;
	    }
	}
    }
    else {
	/* keep track of names for the matrix of names feature (option -s). */
	if (!is_existing_variable(name)) 
	    add_new_variable(name);

	if (num_return_args) {
	    if (current_arg+1 > num_return_args)
		return TRUE;	/* not an error technically */
	    else if (!(return_args[current_arg] = mxCreateFull(r, c, REAL))) {
		fprintf(stderr, "Could not create return argument for `%s'\n",
			name);
		return FALSE;
	    }
	    else {
		mxSetName(return_args[current_arg], name);
		mxSetPr(return_args[current_arg], yp);
		current_arg++;
	    }
	}
	else {
	    if (mexPutFull(name, r, c, yp, NULL) != 0) {
		fprintf(stderr, "Could not create variable for `%s'\n",
			name);
		return FALSE;
	    }
	}
    }
    
    DBG(fprintf(stderr, "exiting intern\n"));
    return TRUE;
}

/**
   Add into the Matlab workspace a vector of strings named #name#. Each
   string is separated by a #SEPARATOR# character (see the defines in this
   file). Otherwise this function operates just like #intern(...)#.

   @param name Name of the variable. 
   @param s Pointer to the string data
   @param extend If TRUE extend the existing variable #name# */

int
intern_strings(char *name, char *s, int extend)
{
    Matrix *mp;
    int status = TRUE;

    name = enforce_name_limits(name);

    DBG(fprintf(stderr,
		"entering intern_string, name: %s, s: %s, ext: %d\n",
		name, s, extend));

    if (extend && is_existing_variable(name)) {
	fprintf(stderr, "extending %s\n", name);
	if (!extend_strings(name, s)) {
	    fprintf(stderr, "Could not create variable for `%s'\n",
		    name);
	    return FALSE;
	}
    }

    if (!is_existing_variable(name)) 
	add_new_variable(name);

    if (num_return_args) {
	if (current_arg+1 > num_return_args)
	    return TRUE;	/* not an error technically */

	return_args[current_arg] = mxCreateString(s);
	mxSetName(return_args[current_arg], name);
	current_arg++;
    }
    else {
	mp = mxCreateString(s);
	if (mp)
	    mxSetName(mp, name);
	status = mp && mexPutMatrix(mp) == 0;
    }
    
    DBG(fprintf(stderr, "exiting intern_strings\n"));

    return status;
}

#ifdef TEST

/* 
   Load the .mat file extend_test.mat and run `extend' to test this code.
*/

void 
mexFunction(int nlhs, Matrix *plhs[], int nrhs, Matrix *prhs[])
{
    double data[2][3]={{7.0, 8.0, 9.0}, {10.0, 11.0, 12.0}};
    double data2[2][2]={{7.0, 8.0}, {9.0, 10.0}};
    double vec[3]={2.0, 6.0, 9.0};

    extend_array("testa", 2, 3, &data[0][0]);

    extend_array("testa2", 2, 2, &data2[0][0]);

    extend_array("testv", 3, 1, &vec[0]);

    extend_array("testv2", 3, 1, &vec[0]);
}

#endif
