
/* 
   (c) COPYRIGHT URI/MIT 1996,1998
   Please read the full copyright statement in the file COPYRIGHT.

   Authors:
      jhrg,jimg		James Gallagher (jgallagher@gso.uri.edu)
*/

/** Simple ADT for named vectors. These vectors are dynamically sized so that
    elements can be added one at a time without performance penalties. This is
    intended for use with the DODS/Matlab and DODS/IDL interface software. I
    wrote it because it appears that Matlab 4 provides no easy and efficient
    way to build up vectors a single element at a time.

    Note that although vectors are added to a single value at a time storage
    is allocated in CLUMP bytes at a time. Thus many values can be added with
    only a few allocations.

    @author jhrg */

/* 
 * $Log: variable.c,v $
 * Revision 1.9  1999/04/30 17:06:58  jimg
 * Merged with no-gnu and release-2-24
 *
 * Revision 1.8.14.1  1999/04/29 18:48:30  jimg
 * Fixed comments and copyright.
 *
 * Revision 1.8  1998/09/16 23:03:43  jimg
 * Fixed the mxCalloc/allocator prototype.
 *
 * Revision 1.7  1998/03/21 00:18:58  jimg
 * Added some documentation.
 * Added reorganize_strings() which simplifies interning vectors of strings.
 *
 * Revision 1.6  1998/03/20 00:34:09  jimg
 * Moved many defines to a new header file: defines.h
 *
 * Revision 1.5  1997/12/16 01:01:28  jimg
 * Fixed bug where max_string_len was set to an incorrect value.
 * Moved structure definition to header file.
 *
 * Revision 1.4  1996/12/18 00:51:32  jimg
 * Added include for dmalloc.
 *
 * Revision 1.3  1996/10/29 21:44:43  jimg
 * Added string.h to included files
 *
 * Revision 1.2  1996/10/25 23:04:41  jimg
 * Added debugging statements for malloc/realloc/free.
 * Added support for string vectors.
 * Removed assert.h since that breaks compilation with gcc - gcc will reference
 * `__eprintf' which cmex cannot find.
 *
 * Revision 1.1  1996/10/23 23:55:06  jimg
 * Created.
 */

static char id[]={"$Id: variable.c,v 1.9 1999/04/30 17:06:58 jimg Exp $"};

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif 

/** Allocate storage #CLUMP# bytes at a time. */
#define CLUMP 1000

/** Use this to separate strings in the character vector hack for Matlab 4.
 */
#define SEPARATOR "\n"

#include "defines.h"
#include "variable.h"

/** Search the list VARS for a variable named NAME. If found, return a pointer
    to the element, otherwise return NULL. 

    @see variable */

variable *
find_var(variable *vars, char *name)
{
    while (vars) {
	if (strcmp(vars->name, name) == 0)
	    return vars;
	else
	    vars = vars->next;
    }

    return NULL;
}

/** Add a variable named NAME to the linked list of variables VARS. Allocate
    enough memory to hold CLUMP elements and set the size of the vector to 0.
    The new element is added to the list; A pointer to the new element (and
    the new list head) is returned.

    @return A pointer to the new element. */

variable *
add_var(variable *vars, char *name, type t)
{
    variable *var;

    /* Make new element. */
    var = (variable *)malloc(sizeof(variable));
    DBG2(fprintf(stderr, "malloc: %x (%d bytes)\n", var, sizeof(variable)));
    var->name = (char *)malloc(sizeof(char) * (strlen(name) + 1));
    DBG2(fprintf(stderr, "malloc: %x (%d bytes)\n", var->name, \
		 (strlen(name) + 1)));
    strcpy(var->name, name);
    var->type = t;
    var->length = 0;
    var->max_string_len = 0;

    switch (t) {
      case float64:
	var->data = (double *)malloc(sizeof(double) * CLUMP);
	DBG2(fprintf(stderr, "malloc: %x (%d bytes)\n", var->data, \
		     sizeof(double) * CLUMP));
	break;

      case string:
	var->sdata = (char **)malloc(sizeof(char *) * CLUMP);
	DBG2(fprintf(stderr, "malloc: %x (%d bytes)\n", var->sdata, \
		     sizeof(char *) * CLUMP));
	break;

      default:
	break;
    }
    
    /* Prepend the new element. */
    var->next = vars;

    return var;
}

/** Add #value# onto the end of the vector of values associated with the
    variable #var#.

    @return FALSE if #var# is NULL, TRUE otherwise. */

int
add_float64_value(variable *var, double value)
{
    if (!var || var->type != float64)
	return FALSE;

    if (var->length >= CLUMP) {
	var->data = (double *)realloc(var->data, 
				      sizeof(double) * (var->length + CLUMP));
	DBG2(fprintf(stderr, "realloc: %x (%d bytes)\n", var->data, \
		     sizeof(double) * (var->length + CLUMP)));
    }

    var->data[var->length++] = value;

    return TRUE;
}

/** Add #value# to the vector pointed to by #var#. Update #var#'s
    #max_string_length# field if this is the longest string added so far.

    @return FALSE if #var# is NULL, TRUE otherwise. */

int
add_string_value(variable *var, char *value)
{
    int string_len;

    if (!var || var->type != string)
	return FALSE;

    if (var->length >= CLUMP) {
	var->sdata = (char **)realloc(var->sdata, 
				      sizeof(char *) * (var->length + CLUMP));
	DBG2(fprintf(stderr, "realloc: %x (%d bytes)\n", var->sdata, \
		     sizeof(char *) * (var->length + CLUMP)));
    }

    string_len = strlen(value);
    if (string_len > var->max_string_len)
	var->max_string_len = string_len;

    var->sdata[var->length] = (char *)malloc(string_len + 1);
    strcpy(var->sdata[var->length++], value);

    return TRUE;
}

/** Take the string data in #var# and transform it from a C vector of strings
    to a single vector of characters where individual strings are separated
    by #SEPARATOR# characters.

    @return The new character vector in newly allocated space (using
    mxCalloc()). */

char *
reorganize_strings(variable *var, void *(allocator)(unsigned, unsigned))
{
    char *return_data, *rd;
    int count = var->length;
    int total_length = 0;

    do {
	total_length += strlen(var->sdata[--count]);
    } while (count);

    /* Allocate space for the the chars, separators and a null. rd remembers
       where the array starts for use as the return value */
    rd = return_data = (char *)allocator(total_length + var->length + 1, 1);

    /* copy the strings into the character vector. Go forwards to preserve
       order. */
    count = 0;
    while (count < var->length) {
	DBG(fprintf(stderr, "Adding `%s'\n", var->sdata[count]));
	strcpy(return_data, var->sdata[count]);
	strcat(return_data, SEPARATOR);
	DBG(fprintf(stderr, "return_data: %s\n", rd));
	return_data += strlen(var->sdata[count++]) + 1;
    }

    return rd;
}

/**
   Free the list of variables pointed to by VARS. */

void
free_vars(variable *vars)
{
    variable *v = vars;

    if (!vars)
	return;

    do {
	vars = vars->next;
	DBG2(fprintf(stderr, "free: %x\n", v->name));
	free(v->name);
	switch (v->type) {
	  case float64:
	    DBG2(fprintf(stderr, "free: %x\n", v->data));
	    free(v->data);
	    break;

	  case string:
	    while (v->length--) {
		DBG2(fprintf(stderr, "free: %x\n", v->sdata[v->length]));
		free(v->sdata[v->length]);
	    }
	    DBG2(fprintf(stderr, "free: %x\n", v->sdata));
	    free(v->sdata);
	    break;
	    
	  default:
	    break;
	}

	DBG2(fprintf(stderr, "free: %x\n", v));
	free(v);
	v = vars;
    } while (vars);
}


